#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
from utils import initialise_tests
import numpy as np
import math
import re
import io
import sys
from cil.framework import AcquisitionGeometry, ImageGeometry, BlockGeometry, AcquisitionData
from cil.framework.framework import SystemConfiguration
from cil.framework import Partitioner

initialise_tests()

class Test_AcquisitionGeometry(unittest.TestCase):
    def test_create_Parallel2D(self):

        #default
        AG = AcquisitionGeometry.create_Parallel2D()
        np.testing.assert_allclose(AG.config.system.ray.direction, [0,1], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, [0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, [1,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, [0,0], rtol=1E-6)

        #values
        ray_direction = [0.1, 3.0]
        detector_position = [-1.3,1000.0]
        detector_direction_x = [1,0.2]
        rotation_axis_position = [0.1,2]

        AG = AcquisitionGeometry.create_Parallel2D(ray_direction, detector_position, detector_direction_x, rotation_axis_position)

        ray_direction = np.asarray(ray_direction)
        detector_direction_x = np.asarray(detector_direction_x)

        ray_direction /= np.sqrt((ray_direction**2).sum())
        detector_direction_x /= np.sqrt((detector_direction_x**2).sum())

        np.testing.assert_allclose(AG.config.system.ray.direction, ray_direction, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, detector_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, detector_direction_x, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, rotation_axis_position, rtol=1E-6)

    def test_create_Parallel3D(self):

        #default
        AG = AcquisitionGeometry.create_Parallel3D()
        np.testing.assert_allclose(AG.config.system.ray.direction, [0,1,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, [0,0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, [1,0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_y, [0,0,1], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, [0,0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.direction, [0,0,1], rtol=1E-6)

        #values
        ray_direction = [0.1, 3.0, 0.2]
        detector_position = [-1.3,1000.0, -1.0]
        detector_direction_x = [1,0.2, 0]
        detector_direction_y = [0.0,0,1]
        rotation_axis_position=[0.1, 2,-0.4]
        rotation_axis_direction=[-0.1,-0.3,1]

        AG = AcquisitionGeometry.create_Parallel3D(ray_direction, detector_position, detector_direction_x,detector_direction_y, rotation_axis_position,rotation_axis_direction)

        ray_direction = np.asarray(ray_direction)
        detector_direction_x = np.asarray(detector_direction_x)
        detector_direction_y = np.asarray(detector_direction_y)
        rotation_axis_direction = np.asarray(rotation_axis_direction)

        ray_direction /= np.sqrt((ray_direction**2).sum())
        detector_direction_x /= np.sqrt((detector_direction_x**2).sum())
        detector_direction_y /= np.sqrt((detector_direction_y**2).sum())
        rotation_axis_direction /= np.sqrt((rotation_axis_direction**2).sum())

        np.testing.assert_allclose(AG.config.system.ray.direction, ray_direction, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, detector_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, detector_direction_x, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_y, detector_direction_y, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, rotation_axis_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.direction, rotation_axis_direction, rtol=1E-6)


    def test_create_Cone2D(self):
        #default
        source_position = [0.1, -500.0]
        detector_position = [-1.3,1000.0]

        AG = AcquisitionGeometry.create_Cone2D(source_position, detector_position)
        np.testing.assert_allclose(AG.config.system.source.position, source_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, detector_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, [1,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, [0,0], rtol=1E-6)

        #values
        detector_direction_x = [1,0.2]
        rotation_axis_position = [0.1,2]

        AG = AcquisitionGeometry.create_Cone2D(source_position, detector_position, detector_direction_x, rotation_axis_position)

        detector_direction_x = np.asarray(detector_direction_x)
        detector_direction_x /= np.sqrt((detector_direction_x**2).sum())

        np.testing.assert_allclose(AG.config.system.source.position, source_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, detector_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, detector_direction_x, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, rotation_axis_position, rtol=1E-6)

    def test_create_Cone3D(self):

        #default
        source_position = [0.1, -500.0,-2.0]
        detector_position = [-1.3,1000.0, -1.0]

        AG = AcquisitionGeometry.create_Cone3D(source_position, detector_position)
        np.testing.assert_allclose(AG.config.system.source.position, source_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, detector_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, [1,0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_y, [0,0,1], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, [0,0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.direction, [0,0,1], rtol=1E-6)

        #values
        detector_direction_x = [1,0.2, 0]
        detector_direction_y = [0.0,0,1]
        rotation_axis_position=[0.1, 2,-0.4]
        rotation_axis_direction=[-0.1,-0.3,1]

        AG = AcquisitionGeometry.create_Cone3D(source_position, detector_position, detector_direction_x,detector_direction_y, rotation_axis_position,rotation_axis_direction)

        detector_direction_x = np.asarray(detector_direction_x)
        detector_direction_y = np.asarray(detector_direction_y)
        rotation_axis_direction = np.asarray(rotation_axis_direction)

        detector_direction_x /= np.sqrt((detector_direction_x**2).sum())
        detector_direction_y /= np.sqrt((detector_direction_y**2).sum())
        rotation_axis_direction /= np.sqrt((rotation_axis_direction**2).sum())

        np.testing.assert_allclose(AG.config.system.source.position, source_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, detector_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, detector_direction_x, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_y, detector_direction_y, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, rotation_axis_position, rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.direction, rotation_axis_direction, rtol=1E-6)


    def test_shift_detector_origin_bottom_left(self):
        initial_position = np.array([2.5, -1.3, 10.2])
        pixel_size = np.array([0.5, 0.7])
        detector_direction_x=np.array([1/math.sqrt(2),0,1/math.sqrt(2)])
        detector_direction_y=np.array([-1/math.sqrt(2),0,1/math.sqrt(2)])

        geometry = AcquisitionGeometry.create_Parallel3D(detector_position=initial_position, detector_direction_x=detector_direction_x, detector_direction_y=detector_direction_y)\
                                      .set_panel([10, 5], [0.5, 0.7], origin='bottom-left')\
                                      .set_angles([0])
        # Test horizontal shift to the left
        shift = -1.5
        geometry.config.shift_detector_in_plane(shift, 'horizontal')
        updated_position = geometry.config.system.detector.position
        expected_position = initial_position - detector_direction_x * pixel_size[0] * shift
        np.testing.assert_array_almost_equal(updated_position, expected_position)

        # Test horizontal shift to the right
        shift = 3.0
        geometry.config.shift_detector_in_plane(shift, 'horizontal')
        updated_position = geometry.config.system.detector.position
        expected_position = expected_position - detector_direction_x * pixel_size[0] * shift
        np.testing.assert_array_almost_equal(updated_position, expected_position)

        # Test vertical shift down
        shift = -1.5
        geometry.config.shift_detector_in_plane(shift, 'vertical')
        updated_position = geometry.config.system.detector.position
        expected_position = expected_position - detector_direction_y * pixel_size[1] * shift
        np.testing.assert_array_almost_equal(updated_position, expected_position)

        # Test vertical shift up
        shift = 3.0
        geometry.config.shift_detector_in_plane(shift, 'vertical')
        updated_position = geometry.config.system.detector.position
        expected_position = expected_position - detector_direction_y * pixel_size[1] * shift
        np.testing.assert_array_almost_equal(updated_position, expected_position)


    def test_shift_detector_origin_top_right(self):
        initial_position = np.array([2.5, -1.3, 10.2])
        detector_direction_x=np.array([1/math.sqrt(2),0,1/math.sqrt(2)])
        detector_direction_y=np.array([-1/math.sqrt(2),0,1/math.sqrt(2)])

        pixel_size = np.array([0.5, 0.7])
        geometry = AcquisitionGeometry.create_Parallel3D(detector_position=initial_position, detector_direction_x=detector_direction_x, detector_direction_y=detector_direction_y)\
                                      .set_panel([10, 5], [0.5, 0.7], origin='top-right')\
                                      .set_angles([0])

        # Test horizontal shift to the right
        shift = -1.5
        geometry.config.shift_detector_in_plane(shift, 'horizontal')
        updated_position = geometry.config.system.detector.position
        expected_position = initial_position + detector_direction_x * pixel_size[0] * shift
        np.testing.assert_array_almost_equal(updated_position, expected_position)

        # Test horizontal shift to the left
        shift = 3.0
        geometry.config.shift_detector_in_plane(shift, 'horizontal')
        updated_position = geometry.config.system.detector.position
        expected_position = expected_position + detector_direction_x * pixel_size[0] * shift
        np.testing.assert_array_almost_equal(updated_position, expected_position)

        # Test vertical shift up
        shift = -1.5
        geometry.config.shift_detector_in_plane(shift, 'vertical')
        updated_position = geometry.config.system.detector.position
        expected_position = expected_position + detector_direction_y * pixel_size[1] * shift
        np.testing.assert_array_almost_equal(updated_position, expected_position)

        # Test vertical shift down
        shift = 3.0
        geometry.config.shift_detector_in_plane(shift, 'vertical')
        updated_position = geometry.config.system.detector.position
        expected_position = expected_position +  detector_direction_y * pixel_size[1] * shift
        np.testing.assert_array_almost_equal(updated_position, expected_position)


    def test_SystemConfiguration(self):

        #SystemConfiguration error handeling
        AG = AcquisitionGeometry.create_Parallel3D()

        #vector wrong length
        with self.assertRaises(ValueError):
            AG.config.system.detector.position = [-0.1,0.1]

        #detector xs and yumns should be perpendicular
        with self.assertRaises(ValueError):
            AG.config.system.detector.set_direction([1,0,0],[-0.1,0.1,1])

    def test_set_angles(self):

        AG = AcquisitionGeometry.create_Parallel2D()
        angles = np.linspace(0, 360, 10, dtype=np.float32)

        #default
        AG.set_angles(angles)
        np.testing.assert_allclose(AG.config.angles.angle_data, angles, rtol=1E-6)
        self.assertEqual(AG.config.angles.initial_angle, 0.0)
        self.assertEqual(AG.config.angles.angle_unit, 'degree')

        #values
        AG.set_angles(angles, 0.1, 'radian')
        np.testing.assert_allclose(AG.config.angles.angle_data, angles, rtol=1E-6)
        self.assertEqual(AG.config.angles.initial_angle, 0.1)
        self.assertEqual(AG.config.angles.angle_unit, 'radian')

    def test_set_panel(self):
        AG = AcquisitionGeometry.create_Parallel3D()

        #default
        AG.set_panel([1000,2000])
        np.testing.assert_array_equal(AG.config.panel.num_pixels, [1000,2000])
        np.testing.assert_array_almost_equal(AG.config.panel.pixel_size, [1,1])

        #values
        AG.set_panel([1000,2000],[0.1,0.2])
        np.testing.assert_array_equal(AG.config.panel.num_pixels, [1000,2000])
        np.testing.assert_array_almost_equal(AG.config.panel.pixel_size, [0.1,0.2])

        #set 2D panel with 3D geometry
        with self.assertRaises(ValueError):
            AG.config.panel.num_pixels = [5]

    def test_set_channels(self):
        AG = AcquisitionGeometry.create_Parallel2D()

        #default
        AG.set_channels()
        self.assertEqual(AG.config.channels.num_channels, 1)

        #values
        AG.set_channels(3, ['r','g','b'])
        self.assertEqual(AG.config.channels.num_channels, 3)
        self.assertEqual(AG.config.channels.channel_labels, ['r','g','b'])

        #set wrong length list
        with self.assertRaises(ValueError):
            AG.config.channels.channel_labels = ['a']

    def test_set_labels(self):

        AG = AcquisitionGeometry.create_Parallel3D()
        AG.set_channels(4)
        AG.set_panel([2,3])
        AG.set_angles([0,1,2,3,5])

        #default
        self.assertEqual(AG.dimension_labels, ('channel','angle','vertical','horizontal'))
        self.assertEqual(AG.shape, (4,5,3,2))

        #values
        AG.set_angles([0])
        AG.set_labels(('horizontal','channel','vertical'))
        self.assertEqual(AG.dimension_labels, ('horizontal','channel','vertical'))
        self.assertEqual(AG.shape, (2,4,3))

    def test_get_centre_of_rotation(self):

        # Functionality is tested in specific implementations
        # this checks the pixel size scaling and return format for each geometry type

        gold1_2D = {'offset':(0.25,'units distance'), 'angle':(0.0,'radian')}
        gold2_2D = {'offset':(0.5,'pixels'), 'angle':(0.0,'degree')}
        gold1_3D = {'offset':(0.25,'units distance'), 'angle':(math.pi/4,'radian')}
        gold2_3D = {'offset':(0.5,'pixels'), 'angle':(45,'degree')}

        #check outputs for each geometry type
        ag = AcquisitionGeometry.create_Parallel2D(rotation_axis_position=[0.25, 0.0]).set_panel(10,0.5)
        out1 = ag.get_centre_of_rotation()
        out2 = ag.get_centre_of_rotation(distance_units='pixels', angle_units='degree')
        self.assertDictEqual(gold1_2D, out1, "Failed Parallel2D")
        self.assertDictEqual(gold2_2D, out2, "Failed Parallel2D")

        ag = AcquisitionGeometry.create_Parallel3D(rotation_axis_position=[0.25, 0.0, 0.0], rotation_axis_direction=[0.5,0.0,0.5]).set_panel([10,10],[0.5,0.5])
        out1 = ag.get_centre_of_rotation()
        out2 = ag.get_centre_of_rotation(distance_units='pixels', angle_units='degree')
        self.assertDictEqual(gold1_3D, out1, "Failed Parallel3D")
        self.assertDictEqual(gold2_3D, out2, "Failed Parallel3D")

        ag = AcquisitionGeometry.create_Cone2D([0,-50], [0,50],rotation_axis_position=[0.125, 0.0]).set_panel(10,0.5)
        out1 = ag.get_centre_of_rotation()
        out2 = ag.get_centre_of_rotation(distance_units='pixels', angle_units='degree')
        self.assertDictEqual(gold1_2D, out1, "Failed Cone2D")
        self.assertDictEqual(gold2_2D, out2, "Failed Cone2D")

        ag = AcquisitionGeometry.create_Cone3D([0,-50,0], [0,50,0], rotation_axis_position=[0.125, 0.0, 0.0], rotation_axis_direction=[0.5,0.0,0.5]).set_panel([10,10],[0.5,0.5])
        out1 = ag.get_centre_of_rotation()
        out2 = ag.get_centre_of_rotation(distance_units='pixels', angle_units='degree')
        self.assertDictEqual(gold1_3D, out1, "Failed Cone3D")
        self.assertDictEqual(gold2_3D, out2, "Failed Cone3D")

        with self.assertRaises(ValueError):
            ag.get_centre_of_rotation(distance_units='bad input')

        with self.assertRaises(ValueError):
            ag.get_centre_of_rotation(angle_units='bad input')


    def test_set_centre_of_rotation(self):
        # Functionality is tested in specific implementations
        # this checks the pixel size scaling and return format for each geometry type

        gold_2D = {'offset':(0.25,'units distance'), 'angle':(0.0,'radian')}
        gold_3D = {'offset':(0.25,'units distance'), 'angle':(math.pi/4,'radian')}

        #check outputs for each geometry type
        ag = AcquisitionGeometry.create_Parallel2D().set_panel(10,0.5)
        ag.set_centre_of_rotation(0.25)
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_2D, out, "Failed Parallel2D default")

        ag.set_centre_of_rotation(0.5, 'pixels')
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_2D, out, "Failed Parallel2D unit")

        ag = AcquisitionGeometry.create_Parallel3D().set_panel([10,10],[0.5,0.5])
        ag.set_centre_of_rotation(0.25, angle=math.pi/4)
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_3D, out, "Failed Parallel3D default")

        ag.set_centre_of_rotation(0.5, 'pixels', 45, 'degree')
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_3D, out, "Failed Parallel3D units")

        ag = AcquisitionGeometry.create_Cone2D([0,-50], [0,50]).set_panel(10,0.5)
        ag.set_centre_of_rotation(0.25)
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_2D, out, "Failed Cone2D default")

        ag.set_centre_of_rotation(0.5, 'pixels')
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_2D, out, "Failed Cone2D units")

        ag = AcquisitionGeometry.create_Cone3D([0,-50,0], [0,50,0]).set_panel([10,10],[0.5,0.5])
        ag.set_centre_of_rotation(0.25, angle=math.pi/4)
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_3D, out, "Failed Cone3D default")

        ag.set_centre_of_rotation(0.5,'pixels', 45, 'degree')
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_3D, out, "Failed Cone3D units")

        with self.assertRaises(ValueError):
            ag.set_centre_of_rotation(distance_units='bad input')

        with self.assertRaises(ValueError):
            ag.set_centre_of_rotation(angle_units='bad input')


    def test_set_centre_of_rotation_by_slice(self):
        # Functionality is tested in specific implementations
        # this checks the pixel size scaling and return format for each geometry type

        gold_2D = {'offset':(0.25,'units distance'), 'angle':(0.0,'radian')}
        gold_3D = {'offset':(0.25,'units distance'), 'angle':(math.pi/4,'radian')}

        #check outputs for each geometry type
        ag = AcquisitionGeometry.create_Parallel2D().set_panel(10,0.5)
        ag.set_centre_of_rotation_by_slice(0.5)
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_2D, out, "Failed Parallel2D")

        ag = AcquisitionGeometry.create_Parallel3D().set_panel([10,10],[0.5,0.5])
        ag.set_centre_of_rotation_by_slice(-4.5, -5, 5.5, 5)
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_3D, out, "Failed Parallel3D")

        ag = AcquisitionGeometry.create_Cone2D([0,-50], [0,50]).set_panel(10,0.5)
        ag.set_centre_of_rotation_by_slice(0.5)
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_2D, out, "Failed Cone2D")

        ag = AcquisitionGeometry.create_Cone3D([0,-50,0], [0,50,0]).set_panel([10,10],[0.5,0.5])
        ag.set_centre_of_rotation_by_slice(-4.5, -5, 5.5, 5)
        out = ag.get_centre_of_rotation()
        self.assertDictEqual(gold_3D, out, "Failed Cone3D")


    def test_equal(self):
        AG = AcquisitionGeometry.create_Parallel3D()
        AG.set_channels(4, ['a','b','c','d'])
        AG.set_panel([2,3])
        AG.set_angles([0,1,2,3,5])
        AG.set_labels(('horizontal','angle','vertical','channel'))

        AG2 = AcquisitionGeometry.create_Parallel3D()
        AG2.set_channels(4, ['a','b','c','d'])
        AG2.set_panel([2,3])
        AG2.set_angles([0,1,2,3,5])
        AG2.set_labels(('horizontal','angle','vertical','channel'))
        self.assertTrue(AG == AG2)

        #test not equal
        AG3 = AG2.copy()
        AG3.config.system.ray.direction = [1,0,0]
        self.assertFalse(AG == AG3)

        AG3 = AG2.copy()
        AG3.config.panel.num_pixels = [1,2]
        self.assertFalse(AG == AG3)

        AG3 = AG2.copy()
        AG3.config.channels.channel_labels = ['d','b','c','d']
        self.assertFalse(AG == AG3)

        AG3 = AG2.copy()
        AG3.config.angles.angle_unit ='radian'
        self.assertFalse(AG == AG3)

        AG3 = AG2.copy()
        AG3.config.angles.angle_data[0] = -1
        self.assertFalse(AG == AG3)

    def test_clone(self):
        AG = AcquisitionGeometry.create_Parallel3D()
        AG.set_channels(4)
        AG.set_panel([2,3])
        AG.set_angles([0,1,2,3,5])
        AG.set_labels(('horizontal','angle','vertical','channel'))

        AG2 = AG.clone()
        self.assertEqual(AG2, AG)

    def test_copy(self):
        AG = AcquisitionGeometry.create_Parallel3D()
        AG.set_channels(4)
        AG.set_panel([2,3])
        AG.set_angles([0,1,2,3,5])
        AG.set_labels(('horizontal','angle','vertical','channel'))

        AG2 = AG.copy()
        self.assertEqual(AG2, AG)

    def test_get_centre_slice(self):
        AG = AcquisitionGeometry.create_Parallel3D(detector_direction_y=[0,1,1])
        AG.set_panel([1000,2000],[1,1])
        AG_cs = AG.get_centre_slice()

        AG2 = AcquisitionGeometry.create_Parallel2D()
        AG2.set_panel([1000,1],[1,math.sqrt(0.5)])

        self.assertEqual(AG2, AG_cs)

    def test_allocate(self):
        AG = AcquisitionGeometry.create_Parallel3D()
        AG.set_channels(4)
        AG.set_panel([2,3])
        AG.set_angles([0,1,2,3,5])
        AG.set_labels(('horizontal','angle','vertical','channel'))

        test = AG.allocate()
        test2 = np.ndarray([2,5,3,4])
        self.assertEqual(test.shape, test2.shape)

    def test_get_ImageGeometry(self):

        AG = AcquisitionGeometry.create_Parallel2D()\
            .set_panel(num_pixels=[512,1],pixel_size=[0.1,0.1])
        IG = AG.get_ImageGeometry()
        IG_gold = ImageGeometry(512,512,0,0.1,0.1,1,0,0,0,1)
        self.assertEqual(IG, IG_gold)

        AG = AcquisitionGeometry.create_Parallel3D()\
            .set_panel(num_pixels=[512,3],pixel_size=[0.1,0.2])
        IG = AG.get_ImageGeometry()
        IG_gold = ImageGeometry(512,512,3,0.1,0.1,0.2,0,0,0,1)
        self.assertEqual(IG, IG_gold)

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,500.])\
            .set_panel(num_pixels=[512,1],pixel_size=[0.1,0.2])
        IG = AG.get_ImageGeometry()
        IG_gold = ImageGeometry(512,512,0,0.05,0.05,1,0,0,0,1)
        self.assertEqual(IG, IG_gold)

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,500.,0])\
            .set_panel(num_pixels=[512,3],pixel_size=[0.1,0.2])
        IG = AG.get_ImageGeometry()
        IG_gold = ImageGeometry(512,512,3,0.05,0.05,0.1,0,0,0,1)
        self.assertEqual(IG, IG_gold)

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,500.,0])\
            .set_panel(num_pixels=[512,3],pixel_size=[0.1,0.2])
        IG = AG.get_ImageGeometry(resolution=0.5)
        IG_gold = ImageGeometry(256,256,2,0.1,0.1,0.2,0,0,0,1)
        self.assertEqual(IG, IG_gold)

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,500.,0])\
            .set_panel(num_pixels=[512,3],pixel_size=[0.1,0.2])
        IG = AG.get_ImageGeometry(resolution=2)
        IG_gold = ImageGeometry(1024,1024,6,0.025,0.025,0.05,0,0,0,1)
        self.assertEqual(IG, IG_gold)


class AlignGeometries(unittest.TestCase):

    def test_set_origin(self):

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-50,0], detector_position=[0.,100.,0])\
            .set_panel(num_pixels=[50,50])

        self.assertTrue(True)


    def test_rotation_vec_to_y(self):

        M = SystemConfiguration.rotation_vec_to_y([0,1])
        a = np.array([[1, 0],[0, 1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([0,-1])
        a = np.array([[-1, 0],[0, -1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([1,1])
        a = np.array([[0.70710678, -0.70710678],[0.70710678, 0.70710678]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([1,-1])
        a = np.array([[-0.70710678, -0.70710678],[0.70710678, -0.70710678]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([0,1,0])
        a = np.array([[1, 0, 0],[0, 1, 0],[0, 0, 1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([0,-1,0])
        a = np.array([[-1, 0, 0],[0, -1, 0],[0, 0, 1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([0,1,1])
        a = np.array([[1, 0, 0],[0, 1, 0],[0, 0, 1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([0,-1,1])
        a = np.array([[-1, 0, 0],[0, -1, 0],[0, 0, 1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([1,1,0])
        a = np.array([[0.70710678, -0.70710678, 0],[0.70710678, 0.70710678, 0],[0, 0, 1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([1,-1,0])
        a = np.array([[-0.70710678, -0.70710678, 0],[0.70710678, -0.70710678, 0],[0, 0, 1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_y([1,-1,1])
        a = np.array([[-0.70710678, -0.70710678, 0],[0.70710678, -0.70710678, 0],[0, 0, 1]])
        np.testing.assert_allclose(M,a, atol=1e-6)


    def test_rotation_vec_z(self):

        M = SystemConfiguration.rotation_vec_to_z([0,0,1])
        a = np.array([[1, 0, 0],[0, 1, 0],[0, 0, 1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_z([0,0,-1])
        a = np.array([[1, 0, 0],[0, -1, 0],[0, 0, -1]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_z([1,0,0])
        a = np.array([[0, 0, -1],[0, 1, 0],[1, 0, 0]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_z([-1,0,0])
        a = np.array([[0, 0, 1],[0, 1, 0],[-1, 0, 0]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_z([0,1,0])
        a = np.array([[1, 0, 0],[0, 0, -1],[0, 1, 0]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_z([0,-1,0])
        a = np.array([[1, 0, 0],[0, 0, 1],[0, -1, 0]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_z([1,-1,0])
        a = np.array([[0.5, 0.5, -0.70710678],[0.5, 0.5, 0.70710678],[0.70710678, -0.70710678, 0]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_z([1,0,1])
        a = np.array([[0.70710678, 0, -0.70710678],[0,1,0],[0.70710678, 0, 0.70710678]])
        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_z([0,1,-1])
        a = np.array([[1,0,0],[0, -0.70710678, -0.70710678],[0, 0.70710678, -0.70710678]])

        np.testing.assert_allclose(M,a, atol=1e-6)

        M = SystemConfiguration.rotation_vec_to_z([-1,-1,-1])
        a = np.array([[0.21132491, -0.78867509, 0.57735025],[ -0.78867509, 0.21132491, 0.57735025],[-0.57735025, -0.57735025, -0.57735025]])

        np.testing.assert_allclose(M,a, atol=1e-6)


class Test_Parallel2D(unittest.TestCase):


    def test_align_reference_frame_cil(self):

        ag = AcquisitionGeometry.create_Parallel2D(ray_direction=[0,-1], detector_position=[0.,-100.], rotation_axis_position=[10.,5.])
        ag.config.system.align_reference_frame('cil')

        np.testing.assert_allclose(ag.config.system.ray.direction, [0,1], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.position, [10,105], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.direction_x, [-1,0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.rotation_axis.position, [0,0], rtol=1E-6)


    def test_align_reference_frame_tigre(self):

        ag = AcquisitionGeometry.create_Parallel2D(ray_direction=[0,-1], detector_position=[0.,-100.], rotation_axis_position=[10.,5.])
        ag.config.system.align_reference_frame('tigre')

        np.testing.assert_allclose(ag.config.system.ray.direction, [0,1], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.position, [10,105], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.direction_x, [-1,0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.rotation_axis.position, [0,0], rtol=1E-6)


    def test_system_description(self):
        AG = AcquisitionGeometry.create_Parallel2D()
        self.assertTrue(AG.system_description=='simple')

        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[5,0], rotation_axis_position=[5,0])
        self.assertTrue(AG.system_description=='simple')

        AG = AcquisitionGeometry.create_Parallel2D(ray_direction=[1,1])
        self.assertTrue(AG.system_description=='advanced')

        AG = AcquisitionGeometry.create_Parallel2D(rotation_axis_position=[5,0])
        self.assertTrue(AG.system_description=='offset')

    def test_get_centre_slice(self):
        AG = AcquisitionGeometry.create_Parallel2D()
        AG2 = AG.copy()

        AG2.config.system.get_centre_slice()
        self.assertEqual(AG.config.system, AG2.config.system)

    def test_calculate_magnification(self):
        AG = AcquisitionGeometry.create_Parallel2D()
        out = AG.config.system.calculate_magnification()
        self.assertEqual(out, [None, None, 1])

    def test_calculate_centre_of_rotation(self):
        AG = AcquisitionGeometry.create_Parallel2D()
        out = AG.config.system.calculate_centre_of_rotation()
        gold = {'offset':(0,'units')}
        gold = (0,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed basic")


        AG = AcquisitionGeometry.create_Parallel2D(rotation_axis_position=[0.5,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (0.5,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        AG = AcquisitionGeometry.create_Parallel2D(rotation_axis_position=[-0.5,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (-0.5,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        AG = AcquisitionGeometry.create_Parallel2D(rotation_axis_position=[0.5,0.], detector_direction_x=[-1,0])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (-0.5,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed detector direction")

        theta = math.pi/4 #detector angle
        distance = 0.5 / math.cos(theta)
        AG = AcquisitionGeometry.create_Parallel2D(detector_direction_x=[0.5,0.5],rotation_axis_position=[0.5,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (distance,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed with rotated detector")

    def test_set_centre_of_rotation(self):
        AG = AcquisitionGeometry.create_Parallel2D()

        gold = (1.5, 0)
        AG.config.system.set_centre_of_rotation(gold[0])
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        gold = (-1.5, 0)
        AG.config.system.set_centre_of_rotation(gold[0])
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        gold = (0, 0)
        AG.config.system.set_centre_of_rotation(gold[0])
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed reset offset",atol=1e-10)

        AG = AcquisitionGeometry.create_Parallel2D(detector_direction_x=[-1,0])

        offset_in = 1.5
        gold = (1.5, 0)
        AG.config.system.set_centre_of_rotation(gold[0])
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        gold = (-1.5, 0)
        AG.config.system.set_centre_of_rotation(gold[0])
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        gold = (0, 0)
        AG.config.system.set_centre_of_rotation(gold[0])
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg= "Failed reset offset",atol=1e-10)


        AG = AcquisitionGeometry.create_Parallel2D(detector_direction_x=[0.5,0.5],rotation_axis_position=[0.5,0.])
        gold = (1.5, 0)
        AG.config.system.set_centre_of_rotation(gold[0])
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        gold = (-1.5, 0)
        AG.config.system.set_centre_of_rotation(gold[0])
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        offset_in = 0
        gold = (0, 0)
        AG.config.system.set_centre_of_rotation(gold[0])
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed reset offset",atol=1e-10)


class Test_Parallel3D(unittest.TestCase):

    def test_align_reference_frame_cil(self):

        ag = AcquisitionGeometry.create_Parallel3D(ray_direction=[0,-1,0], detector_position=[0.,-100.,0], rotation_axis_position=[10.,5.,0], rotation_axis_direction=[0,0,-1])
        ag.config.system.align_reference_frame('cil')

        np.testing.assert_allclose(ag.config.system.ray.direction, [0,1, 0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.position, [-10,105,0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.direction_x, [1,0,0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.direction_y, [0,0,-1], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.rotation_axis.position, [0,0,0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.rotation_axis.direction, [0,0,1], rtol=1E-6)


    def test_align_reference_frame_tigre(self):

        ag = AcquisitionGeometry.create_Parallel3D(ray_direction=[0,-1,0], detector_position=[0.,-100.,0], rotation_axis_position=[10.,5.,0], rotation_axis_direction=[0,0,-1])
        ag.config.system.align_reference_frame('tigre')

        np.testing.assert_allclose(ag.config.system.ray.direction, [0,1, 0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.position, [-10,105,0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.direction_x, [1,0,0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.detector.direction_y, [0,0,-1], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.rotation_axis.position, [0,0,0], rtol=1E-6)
        np.testing.assert_allclose(ag.config.system.rotation_axis.direction, [0,0,1], rtol=1E-6)


    def test_system_description(self):
        AG = AcquisitionGeometry.create_Parallel3D()
        self.assertTrue(AG.system_description=='simple')

        AG = AcquisitionGeometry.create_Parallel3D(detector_position=[5,0,0], rotation_axis_position=[5,0,0])
        self.assertTrue(AG.system_description=='simple')

        AG = AcquisitionGeometry.create_Parallel3D(rotation_axis_position=[5,0,0])
        self.assertTrue(AG.system_description=='offset')

        AG = AcquisitionGeometry.create_Parallel3D(ray_direction = [1,1,0])
        self.assertTrue(AG.system_description=='advanced')


    def test_get_centre_slice(self):
        #returns the 2D version
        AG = AcquisitionGeometry.create_Parallel3D()
        AG2 = AcquisitionGeometry.create_Parallel2D()
        cs = AG.config.system.get_centre_slice()
        self.assertEqual(cs, AG2.config.system)

        #returns the 2D version
        AG = AcquisitionGeometry.create_Parallel3D(rotation_axis_direction=[-1,0,1],detector_direction_x=[1,0,1], detector_direction_y=[-1,0,1])
        AG2 = AcquisitionGeometry.create_Parallel2D()
        cs = AG.config.system.get_centre_slice()
        self.assertEqual(cs, AG2.config.system)

        #raise error if cannot extract a cnetre slice
        AG = AcquisitionGeometry.create_Parallel3D(ray_direction=[0,1,1])
        with self.assertRaises(ValueError):
            cs = AG.config.system.get_centre_slice()

        AG = AcquisitionGeometry.create_Parallel3D(detector_direction_x=[1,0,1], detector_direction_y=[-1,0,1])
        with self.assertRaises(ValueError):
            cs = AG.config.system.get_centre_slice()

    def test_calculate_magnification(self):
        AG = AcquisitionGeometry.create_Parallel3D()
        out = AG.config.system.calculate_magnification()
        self.assertEqual(out, [None, None, 1])

    def test_calculate_centre_of_rotation(self):

        AG = AcquisitionGeometry.create_Parallel3D()
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (0,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed basic")

        angle = math.pi/4
        AG = AcquisitionGeometry.create_Parallel3D(rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[0.5,0,0.5])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (0.5,angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        AG = AcquisitionGeometry.create_Parallel3D(rotation_axis_position=[-0.5,0.,0.], rotation_axis_direction=[-0.5,0,0.5])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (-0.5,-angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        AG = AcquisitionGeometry.create_Parallel3D(rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[0.5,0,0.5], detector_direction_x=[-1,0,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (-0.5,-angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed detector direction_x")

        AG = AcquisitionGeometry.create_Parallel3D(rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[0.5,0,0.5], detector_direction_y=[0,0,-1])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (0.5,math.pi-angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed detector direction_y")

        AG = AcquisitionGeometry.create_Parallel3D(rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[-0.5,0,-0.5], detector_direction_y=[0,0,-1])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (0.5,-angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed invert rotate axis")


        theta = math.pi/4 #detector angle
        distance = 0.5 / math.cos(theta)
        AG = AcquisitionGeometry.create_Parallel3D(detector_direction_x=[0.5,0.5,0.],rotation_axis_position=[0.5,0.,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = {'offset':(distance,'units')}
        gold = (distance,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed rotated detector")

        AG = AcquisitionGeometry.create_Parallel3D(detector_direction_y=[0.0,-0.5,0.5],rotation_axis_position=[0.5,0.,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (0.5,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed tilted detector")


    def test_set_centre_of_rotation(self):
        AG = AcquisitionGeometry.create_Parallel3D()

        gold = (1.5, 0)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")


        gold = (-1.5, 0)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        gold = (0, 0)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed reset offset")

        gold = (0.0, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed positive angle")

        gold = (0.0, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed negative angle")

        gold = (0, 0)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed reset angle")

        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed combination A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed combination B")



        AG = AcquisitionGeometry.create_Parallel3D(detector_direction_x=[-1,0,0.])

        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert detector x A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert detector x B")


        AG = AcquisitionGeometry.create_Parallel3D(rotation_axis_direction=[0,0,-1])

        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert rotate axis A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert rotate axis B")


        AG = AcquisitionGeometry.create_Parallel3D(detector_direction_y=[0,0,-1])
        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert detector y A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert detector y A")

        AG = AcquisitionGeometry.create_Parallel3D(rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[-0.5,0,-0.5], detector_direction_y=[0,0,-1])
        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed rolled and inverted detector x A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed rolled and inverted detector x B")


        AG = AcquisitionGeometry.create_Parallel3D(detector_direction_x=[0.5,0.5,0],rotation_axis_position=[0.5,0.,0])
        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed rotated detector x A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed rotated detector x B")

        AG = AcquisitionGeometry.create_Parallel3D(detector_direction_y=[0.0,-0.5,0.5],rotation_axis_position=[0.5,0.,0.])
        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed tilted detector x A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed tilted detector x B")

class Test_Cone2D(unittest.TestCase):

    def test_align_reference_frame_cil(self):

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,50], detector_position=[0.,-100.], rotation_axis_position=[5.,2.])
        AG.set_panel(100)

        AG.config.system.align_reference_frame('cil')

        np.testing.assert_allclose(AG.config.system.source.position, [5,-48], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, [5,102], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, [-1,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, [0,0], rtol=1E-6)


    def test_align_reference_frame_tigre(self):

        ag = AcquisitionGeometry.create_Cone2D(source_position=[0,50], detector_position=[0.,-100.], rotation_axis_position=[5.,2])
        ag.set_panel(100)

        ag_align = ag.copy()
        ag_align.config.system.align_reference_frame('tigre')

        np.testing.assert_allclose(ag_align.config.system.source.position, [0,-ag.dist_source_center], atol=1E-6)
        np.testing.assert_allclose(ag_align.config.system.rotation_axis.position, [0,0], rtol=1E-6)

        cos_theta = abs(ag.config.system.source.position[1]-ag.config.system.rotation_axis.position[1])/ ag.dist_source_center
        sin_theta = math.sin(math.acos(cos_theta))

        vec = ag.config.system.detector.position-ag.config.system.source.position
        tmp = abs(vec[1])*cos_theta
        det_y = tmp - ag.dist_source_center
        det_x =np.sqrt(vec[1] ** 2 - tmp **2)

        np.testing.assert_allclose(ag_align.config.system.detector.position, [det_x, det_y], rtol=1E-6)

        dir_x = -ag.config.system.detector.direction_x[0] * cos_theta
        dir_y = ag.config.system.detector.direction_x[0] * sin_theta
        np.testing.assert_allclose(ag_align.config.system.detector.direction_x, [dir_x, dir_y], rtol=1E-6)


    def test_system_description(self):
        AG = AcquisitionGeometry.create_Cone2D(source_position = [0,-50],detector_position=[0,100])
        self.assertTrue(AG.system_description=='simple')

        AG = AcquisitionGeometry.create_Cone2D(source_position = [5,-50],detector_position=[5,100], rotation_axis_position=[5,0])
        self.assertTrue(AG.system_description=='simple')

        AG = AcquisitionGeometry.create_Cone2D(source_position = [5,-50],detector_position=[0,100])
        self.assertTrue(AG.system_description=='advanced')

        AG = AcquisitionGeometry.create_Cone2D(source_position = [0,-50],detector_position=[0,100], rotation_axis_position=[5,0])
        self.assertTrue(AG.system_description=='offset')

    def test_get_centre_slice(self):
        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.])
        AG2 = AG.copy()

        AG2.config.system.get_centre_slice()
        self.assertEqual(AG.config.system, AG2.config.system)

    def test_calculate_magnification(self):
        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.])
        out = AG.config.system.calculate_magnification()
        np.testing.assert_almost_equal(out, [500, 1000, 3])

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.], rotation_axis_position=[0.,250.])
        out = AG.config.system.calculate_magnification()
        np.testing.assert_almost_equal(out, [750, 750, 2])

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.], rotation_axis_position=[5.,0.])
        out = AG.config.system.calculate_magnification()
        source_to_object = np.sqrt(5.0**2 + 500.0**2)
        theta = math.atan2(5.0,500.0)
        source_to_detector = 1500.0/math.cos(theta)
        np.testing.assert_almost_equal(out, [source_to_object, source_to_detector - source_to_object, source_to_detector/source_to_object])

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.], rotation_axis_position=[5.,0.],detector_direction_x=[math.sqrt(5),math.sqrt(5)])
        out = AG.config.system.calculate_magnification()
        source_to_object = np.sqrt(5.0**2 + 500.0**2)

        ab = (AG.config.system.rotation_axis.position - AG.config.system.source.position)/source_to_object

        #source_position + d * ab = detector_position + t * detector_direction_x
        #x: d *  ab[0] =  t * detector_direction_x[0]
        #y: -500 + d *  ab[1] = 1000 + t * detector_direction_x[1]

        # t = (d *  ab[0]) / math.sqrt(5)
        # d = 1500 / (ab[1]  - ab[0])

        source_to_detector = 1500 / (ab[1]  - ab[0])

        np.testing.assert_almost_equal(out, [source_to_object, source_to_detector - source_to_object, source_to_detector/source_to_object])

    def test_calculate_centre_of_rotation(self):
        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (0,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed basic")

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.], rotation_axis_position=[0.5,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (1.5,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.], rotation_axis_position=[-0.5,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (-1.5,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.], rotation_axis_position=[0.5,0.], detector_direction_x=[-1,0])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (-1.5,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed detector direction")

        #offset * mag = 1
        theta = math.pi/4 #detector angle
        phi = math.atan2(1,1000) #ray through rotation axis angle

        L = math.sin(theta)
        X1 = L / math.tan(theta)
        X2 = L / math.tan(math.pi/2 - theta- phi)
        distance = X1 + X2

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,500.], detector_direction_x=[0.5,0.5],rotation_axis_position=[0.5,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (distance,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed with rotated detector")

    def test_set_centre_of_rotation(self):

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.])

        offset_in = 1.5
        AG.config.system.set_centre_of_rotation(offset_in)
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (offset_in,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        offset_in = -1.5
        AG.config.system.set_centre_of_rotation(offset_in)
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (offset_in,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        offset_in = 0
        AG.config.system.set_centre_of_rotation(offset_in)
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (offset_in,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed reset offset", atol=1e-10)

        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,1000.], detector_direction_x=[-1,0])

        offset_in = 1.5
        AG.config.system.set_centre_of_rotation(offset_in)
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (offset_in,0)
        np.testing.assert_allclose(out, gold, err_msg= "Failed positive offset")

        offset_in = -1.5
        AG.config.system.set_centre_of_rotation(offset_in)
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (offset_in,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        offset_in = 0
        AG.config.system.set_centre_of_rotation(offset_in)
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (offset_in,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed reset offset", atol=1e-10)


        AG = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0.,500.], detector_direction_x=[0.5,0.5],rotation_axis_position=[0.5,0.])
        offset_in = 1.5
        AG.config.system.set_centre_of_rotation(offset_in)
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (offset_in,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        offset_in = -1.5
        AG.config.system.set_centre_of_rotation(offset_in)
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (offset_in,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        offset_in = 0
        AG.config.system.set_centre_of_rotation(offset_in)
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (offset_in,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed reset offset", atol=1e-10)


class Test_Cone3D(unittest.TestCase):

    def test_align_reference_frame_cil(self):
        AG = AcquisitionGeometry.create_Cone3D(source_position=[5,500,0],detector_position=[5.,-1000.,0], rotation_axis_position=[5,0,0], rotation_axis_direction=[0,0,-1])
        AG.config.system.align_reference_frame('cil')

        np.testing.assert_allclose(AG.config.system.source.position, [0,-500, 0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, [0,1000,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, [1,0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_y, [0,0,-1], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, [0,0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.direction, [0,0,1], rtol=1E-6)


    def test_align_reference_frame_tigre(self):
        AG = AcquisitionGeometry.create_Cone3D(source_position=[5,500,0],detector_position=[5.,-1000.,0], rotation_axis_position=[5,0,0], rotation_axis_direction=[0,0,-1])
        AG.config.system.align_reference_frame('tigre')

        np.testing.assert_allclose(AG.config.system.source.position, [0,-500, 0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.position, [0,1000,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_x, [1,0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.detector.direction_y, [0,0,-1], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.position, [0,0,0], rtol=1E-6)
        np.testing.assert_allclose(AG.config.system.rotation_axis.direction, [0,0,1], rtol=1E-6)


    def test_system_description(self):
        AG = AcquisitionGeometry.create_Cone3D(source_position = [0,-50,0],detector_position=[0,100,0])
        self.assertTrue(AG.system_description=='simple')

        AG = AcquisitionGeometry.create_Cone3D(source_position = [-50,0,0],detector_position=[100,0,0], detector_direction_x=[0,-1,0])
        self.assertTrue(AG.system_description=='simple')

        AG = AcquisitionGeometry.create_Cone3D(source_position = [5,-50,0],detector_position=[5,100,0], rotation_axis_position=[5,0,0])
        self.assertTrue(AG.system_description=='simple')

        AG = AcquisitionGeometry.create_Cone3D(source_position = [5,-50,0],detector_position=[0,100,0])
        self.assertTrue(AG.system_description=='advanced')

        AG = AcquisitionGeometry.create_Cone3D(source_position = [0,-50,0],detector_position=[0,100,0], rotation_axis_position=[5,0,0])
        self.assertTrue(AG.system_description=='offset')

    def test_get_centre_slice(self):
        #returns the 2D version
        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0,1000,0])
        AG2 = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0,1000])
        cs = AG.config.system.get_centre_slice()
        self.assertEqual(cs, AG2.config.system)

        #returns the 2D version
        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0,1000,0], rotation_axis_direction=[-1,0,1], detector_direction_x=[1,0,1], detector_direction_y=[-1,0,1])
        AG2 = AcquisitionGeometry.create_Cone2D(source_position=[0,-500], detector_position=[0,1000])
        cs = AG.config.system.get_centre_slice()
        self.assertEqual(cs, AG2.config.system)

        #raise error if cannot extract a centre slice
        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0,1000,0], rotation_axis_direction=[1,0,1])
        with self.assertRaises(ValueError):
            cs = AG.config.system.get_centre_slice()

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0,1000,0],detector_direction_x=[1,0,1], detector_direction_y=[-1,0,1])
        with self.assertRaises(ValueError):
            cs = AG.config.system.get_centre_slice()

    def test_calculate_magnification(self):
        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0])
        out = AG.config.system.calculate_magnification()
        self.assertEqual(out, [500, 1000, 3])

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[0.,250.,0])
        out = AG.config.system.calculate_magnification()
        self.assertEqual(out, [750, 750, 2])

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[5.,0.,0])
        out = AG.config.system.calculate_magnification()
        source_to_object = np.sqrt(5.0**2 + 500.0**2)
        theta = math.atan2(5.0,500.0)
        source_to_detector = 1500.0/math.cos(theta)
        self.assertEqual(out, [source_to_object, source_to_detector - source_to_object, source_to_detector/source_to_object])

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[0.,0.,5.])
        out = AG.config.system.calculate_magnification()
        source_to_object = np.sqrt(5.0**2 + 500.0**2)
        theta = math.atan2(5.0,500.0)
        source_to_detector = 1500.0/math.cos(theta)
        self.assertEqual(out, [source_to_object, source_to_detector - source_to_object, source_to_detector/source_to_object])

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0],detector_direction_y=[0,math.sqrt(5),math.sqrt(5)])
        out = AG.config.system.calculate_magnification()
        self.assertEqual(out, [500, 1000, 3])

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0],detector_direction_x=[1,0.1,0.2],detector_direction_y=[-0.2,0,1])
        out = AG.config.system.calculate_magnification()
        self.assertEqual(out, [500, 1000, 3])

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[5.,0.,0],detector_direction_x=[math.sqrt(5),math.sqrt(5),0])
        out = AG.config.system.calculate_magnification()
        source_to_object = np.sqrt(5.0**2 + 500.0**2)

        ab = (AG.config.system.rotation_axis.position - AG.config.system.source.position).astype(np.float64)/source_to_object

        #source_position + d * ab = detector_position + t * detector_direction_x
        #x: d *  ab[0] =  t * detector_direction_x[0]
        #y: -500 + d *  ab[1] = 1000 + t * detector_direction_x[1]

        # t = (d *  ab[0]) / math.sqrt(5)
        # d = 1500 / (ab[1]  - ab[0])

        source_to_detector = 1500 / (ab[1]  - ab[0])
        self.assertEqual(out, [source_to_object, source_to_detector - source_to_object, source_to_detector/source_to_object])

    def test_calculate_centre_of_rotation(self):
        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (0,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed basic")

        angle = math.pi/4
        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[0.5,0,0.5])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (1.5,angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[-0.5,0.,0.], rotation_axis_direction=[-0.5,0,0.5])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (-1.5,-angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[0.5,0,0.5], detector_direction_x=[-1,0,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (-1.5,-angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed detector direction_x")

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[0.5,0,0.5], detector_direction_y=[0,0,-1])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (1.5,math.pi-angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed detector direction_y")

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[-0.5,0,-0.5], detector_direction_y=[0,0,-1])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (1.5,-angle)
        np.testing.assert_allclose(out, gold, err_msg= "Failed invert rotate axis")

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[0,0,-1])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (1.5,math.pi)
        np.testing.assert_allclose(out, gold, err_msg="Failed invert rotate axis")

        #offset * mag = 1
        theta = math.pi/4 #detector angle
        phi = math.atan2(1,1000) #ray through rotation axis angle

        L = math.sin(theta)
        X1 = L / math.tan(theta)
        X2 = L / math.tan(math.pi/2 - theta- phi)
        distance = X1 + X2

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,500.,0], detector_direction_x=[0.5,0.5,0],rotation_axis_position=[0.5,0.,0])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (distance,0)
        np.testing.assert_allclose(out, gold, err_msg="Failed with rotated detector")


        #offset * mag = 1
        theta = math.pi/4 #detector angle
        phi = math.atan2(1,1000) #ray through rotation axis angle
        psi = math.atan2(1,500)
        Y = 2 * math.sin(math.pi/2-psi) / math.sin(math.pi/2-theta+psi)
        L = -Y * math.sin(theta)

        X = L * math.tan(phi)
        angle = math.atan2(X,Y)

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,500.,0], detector_direction_y=[0.0,-0.5,0.5],rotation_axis_position=[0.5,0.,0.])
        out = AG.config.system.calculate_centre_of_rotation()
        gold = (1.0,angle)
        np.testing.assert_allclose(out, gold, err_msg="Failed tilted detector")

    def test_set_centre_of_rotation(self):

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0])

        gold = (1.5, 0)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed positive offset")

        gold = (-1.5, 0)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed negative offset")

        gold = (0, 0)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed reset offset")

        gold = (0.0, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed positive angle")

        gold = (0.0, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed negative angle")

        gold = (0.0, 0.0)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed reset angle")

        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed combination A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed combination A")



        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], detector_direction_x=[-1,0,0.])
        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert detector x A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert detector x B")


        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_direction=[0,0,-1])

        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert rotate axis A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert rotate axis B")


        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], detector_direction_y=[0,0,-1])
        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert detector y A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed invert detector y A")

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,1000.,0], rotation_axis_position=[0.5,0.,0.], rotation_axis_direction=[-0.5,0,-0.5], detector_direction_y=[0,0,-1])
        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed rolled and inverted detector x A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed rolled and inverted detector x A")


        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,500.,0], detector_direction_x=[0.5,0.5,0],rotation_axis_position=[0.5,0.,0])
        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed rotated detector x A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed rotated detector x B")

        AG = AcquisitionGeometry.create_Cone3D(source_position=[0,-500,0], detector_position=[0.,500.,0], detector_direction_y=[0.0,-0.5,0.5],rotation_axis_position=[0.5,0.,0.])
        gold = (-1.5, -0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed tilted detector x A")

        gold = (1.5, 0.2)
        AG.config.system.set_centre_of_rotation(*gold)
        out = AG.config.system.calculate_centre_of_rotation()
        np.testing.assert_allclose(out, gold, err_msg="Failed tilted detector x B")

class TestSubset(unittest.TestCase):
    def setUp(self) -> None:
        return super().setUp()
    def tearDown(self) -> None:
        return super().tearDown()

    def test_partition_indices_int(self):
        par = Partitioner()

        num_batches = 4
        indices = 9
        ret = par._partition_indices(num_batches, indices, stagger=False)
        gold = [[0, 1, 2], [3, 4] ,[5, 6], [7, 8]]

        self.assertListEqual(ret, gold)

        ret = par._partition_indices(num_batches, indices, stagger=True)
        gold = [[0, 4, 8], [1, 5], [2, 6], [3, 7]]

        self.assertListEqual(ret, gold)

    def test_partition_indices_list(self):
        par = Partitioner()

        num_batches = 4
        num_indices = 9
        indices = list(range(num_indices))
        ret = par._partition_indices(num_batches, indices, stagger=False)
        gold = [[0, 1, 2], [3, 4] ,[5, 6], [7, 8]]

        self.assertListEqual(ret, gold)

        ret = par._partition_indices(num_batches, indices, stagger=True)
        gold = [[0, 4, 8], [1, 5], [2, 6], [3, 7]]

        self.assertListEqual(ret, gold)

    def test_AcquisitionData_split_to_BlockGeometry_and_BlockDataContainer(self):
        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[0,10])\
            .set_panel(num_pixels=10)\
            .set_angles(angles=range(9))

        data = AG.allocate(None)
        for i in range(AG.num_projections):
            data.array[i] = i

        self.AcquisitionGeometry_split_to_BlockGeometry(data, 'sequential', 1)
        self.AcquisitionGeometry_split_to_BlockGeometry(data, 'staggered', 1)
        self.AcquisitionGeometry_split_to_BlockGeometry(data, Partitioner.RANDOM_PERMUTATION, 1)

    def test_AcquisitionData_split_to_BlockGeometry_and_BlockDataContainer_2D_order1(self):
        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[0,10])\
            .set_panel(num_pixels=10)\
            .set_angles(angles=range(9))\
            .set_labels(['angle','horizontal'])

        data = AG.allocate(None)
        for i in range(AG.num_projections):
            data.array[i] = i

        self.AcquisitionGeometry_split_to_BlockGeometry(data, 'sequential', 1)
        self.AcquisitionGeometry_split_to_BlockGeometry(data, 'staggered', 1)
        self.AcquisitionGeometry_split_to_BlockGeometry(data, Partitioner.RANDOM_PERMUTATION, 1)

    def test_AcquisitionData_split_to_BlockGeometry_and_BlockDataContainer_2D_order2(self):

        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[0,10])\
            .set_panel(num_pixels=10)\
            .set_angles(angles=range(9))\
            .set_labels(['horizontal', 'angle'])

        data = AG.allocate(None)
        for i in range(AG.num_projections):
            data.array[i] = i

        self.AcquisitionGeometry_split_to_BlockGeometry(data, 'sequential', 1)
        self.AcquisitionGeometry_split_to_BlockGeometry(data, 'staggered', 1)
        self.AcquisitionGeometry_split_to_BlockGeometry(data, Partitioner.RANDOM_PERMUTATION, 1)

    def AcquisitionGeometry_split_to_BlockGeometry(self, data, method, seed):
        num_batches = 4
        np.random.seed(seed)
        datasplit = data.partition(num_batches, method)
        bg = datasplit.geometry
        ag = data.geometry
        num_indices = ag.num_projections

        gold = [ np.zeros(num_indices, dtype=bool) for _ in range(num_batches) ]
        if method == Partitioner.SEQUENTIAL:
            gold = [[0, 1, 2], [3, 4], [5, 6], [7, 8]]

        elif method == Partitioner.STAGGERED:
            gold = [[0, 4, 8], [1, 5], [2, 6], [3, 7]]

        elif method == Partitioner.RANDOM_PERMUTATION:
            # with seed==1
            gold = [[8, 2, 6], [7, 1], [0, 4], [3, 5]]


        for i, geo in enumerate(bg):
            np.testing.assert_allclose(geo.angles, np.asarray(gold[i]))

    def test_geometry_print_angles(self):

        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[0,10])\
            .set_panel(num_pixels=10)\
            .set_angles(angles=range(90))\
            .set_labels(['horizontal', 'angle'])
        AD = AcquisitionData(np.zeros([10,90]), geometry=AG, deep_copy=False)

        # redirect print output
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput

        print(AD.geometry)
        angles = re.findall('Angles [\d]+-[\d]+ in degrees:\s+\[.*\]+', capturedOutput.getvalue(), re.MULTILINE)
        self.assertEqual(angles[0], 'Angles 0-9 in degrees: [0., 1., 2., 3., 4., 5., 6., 7., 8., 9.]')
        self.assertEqual(angles[1], 'Angles 80-89 in degrees: [80., 81., 82., 83., 84., 85., 86., 87., 88., 89.]')

        # test output when angles=31
        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[0,10])\
            .set_panel(num_pixels=10)\
            .set_angles(angles=range(31))\
            .set_labels(['horizontal', 'angle'])
        AD = AcquisitionData(np.zeros([10,31]), geometry=AG, deep_copy=False)
        print(AD.geometry)
        angles = re.findall('Angles [\d]+-[\d]+ in degrees:\s+\[.*\]+', capturedOutput.getvalue(), re.MULTILINE)
        self.assertEqual(angles[2], 'Angles 0-9 in degrees: [0., 1., 2., 3., 4., 5., 6., 7., 8., 9.]')
        self.assertEqual(angles[3], 'Angles 21-30 in degrees: [21., 22., 23., 24., 25., 26., 27., 28., 29., 30.]')

        # test output when angles=30
        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[0,10])\
            .set_panel(num_pixels=10)\
            .set_angles(angles=range(30))\
            .set_labels(['horizontal', 'angle'])
        AD = AcquisitionData(np.zeros([10,30]), geometry=AG, deep_copy=False)
        print(AD.geometry)
        angles = re.findall('Number of positions: 30\n\tAngles [\d]+-[\d]+ in degrees:\s+\[.*\n.*\]+', capturedOutput.getvalue(), re.MULTILINE)
        self.assertEqual(angles[0],\
                'Number of positions: 30\n\tAngles 0-29 in degrees: [ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9., 10., 11., 12., 13., 14.,\n 15., 16., 17., 18., 19., 20., 21., 22., 23., 24., 25., 26., 27., 28., 29.]')

        # test no error occurs when angles<20
        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[0,10])\
            .set_panel(num_pixels=10)\
            .set_angles(angles=range(17))\
            .set_labels(['horizontal', 'angle'])
        AD = AcquisitionData(np.zeros([10,17]), geometry=AG, deep_copy=False)
        print(AD.geometry)

        # test no error occurs when angles<10
        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[0,10])\
            .set_panel(num_pixels=10)\
            .set_angles(angles=range(9))\
            .set_labels(['horizontal', 'angle'])
        AD = AcquisitionData(np.zeros([10,9]), geometry=AG, deep_copy=False)
        print(AD.geometry)

        # test no error occurs when angle=1
        AG = AcquisitionGeometry.create_Parallel2D(detector_position=[0,10])\
            .set_panel(num_pixels=10)\
            .set_angles(angles=range(1))\
            .set_labels(['horizontal', 'angle'])
        AD = AcquisitionData(np.zeros([10,]), geometry=AG, deep_copy=False)
        print(AD.geometry)

        # return to standard print output
        sys.stdout = sys.__stdout__



#  Copyright 2021 United Kingdom Research and Innovation
#  Copyright 2021 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import ImageData, AcquisitionData, AcquisitionGeometry
from cil.framework import DataOrder
from cil.framework.BlockGeometry import BlockGeometry
from cil.optimisation.operators import BlockOperator
from cil.optimisation.operators import LinearOperator
from cil.plugins.tigre import CIL2TIGREGeometry
import numpy as np
import logging
import warnings

log = logging.getLogger(__name__)

try:
    from _Atb import _Atb_ext as Atb
    from _Ax import _Ax_ext as Ax

except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "This plugin requires the additional package TIGRE\n" +
        "Please install it via conda as tigre from the ccpi channel")

try:
    from tigre.utilities.gpu import GpuIds
    has_gpu_sel = True
except ModuleNotFoundError:
    has_gpu_sel = False


class ProjectionOperator(LinearOperator):
    """
        ProjectionOperator configures and calls TIGRE Projectors for your dataset.

        Please refer to the TIGRE documentation for further descriptions
        https://github.com/CERN/TIGRE
        https://iopscience.iop.org/article/10.1088/2057-1976/2/5/055010


        Parameters
        ----------

        image_geometry : `ImageGeometry`, default used if None
            A description of the area/volume to reconstruct

        acquisition_geometry :`AcquisitionGeometry`, `BlockGeometry`
            A description of the acquisition data. If passed a BlockGeometry it will return a BlockOperator.

        direct_method : str,  default 'interpolated'
            The method used by the forward projector, 'Siddon' for ray-voxel intersection, 'interpolated' for interpolated projection

        adjoint_weights : str, default 'matched'
            The weighting method used by the cone-beam backward projector, 'matched' for weights to approximately match the 'interpolated' forward projector, 'FDK' for FDK weights

        Example
        -------
        >>> from cil.plugins.tigre import ProjectionOperator
        >>> PO = ProjectionOperator(image.geometry, data.geometry)
        >>> forward_projection = PO.direct(image)
        >>> backward_projection = PO.adjoint(data)

        """
    def __new__(cls, image_geometry=None, acquisition_geometry=None, \
        direct_method='interpolated',adjoint_weights='matched', **kwargs):
        if isinstance(acquisition_geometry, BlockGeometry):
            log.info("BlockOperator is returned.")

            K = []
            for ag in acquisition_geometry:
                K.append(
                    ProjectionOperator_ag(image_geometry=image_geometry, acquisition_geometry=ag, \
                        direct_method=direct_method, adjoint_weights=adjoint_weights, **kwargs)
                )
            return BlockOperator(*K)
        else:
            log.info("Standard Operator is returned.")
            return super(ProjectionOperator,
                         cls).__new__(ProjectionOperator_ag)


class ProjectionOperator_ag(ProjectionOperator):
    '''TIGRE Projection Operator'''

    def __init__(self,
                 image_geometry=None,
                 acquisition_geometry=None,
                 direct_method='interpolated',
                 adjoint_weights='matched',
                 **kwargs):
        """
        ProjectionOperator configures and calls TIGRE Projectors for your dataset.

        Please refer to the TIGRE documentation for further descriptions
        https://github.com/CERN/TIGRE
        https://iopscience.iop.org/article/10.1088/2057-1976/2/5/055010


        Parameters
        ----------

        image_geometry : ImageGeometry, default used if None
            A description of the area/volume to reconstruct

        acquisition_geometry : AcquisitionGeometry
            A description of the acquisition data

        direct_method : str,  default 'interpolated'
            The method used by the forward projector, 'Siddon' for ray-voxel intersection, 'interpolated' for interpolated projection

        adjoint_weights : str, default 'matched'
            The weighting method used by the cone-beam backward projector, 'matched' for weights to approximately match the 'interpolated' forward projector, 'FDK' for FDK weights

        Example
        -------
        >>> from cil.plugins.tigre import ProjectionOperator
        >>> PO = ProjectionOperator(image.geometry, data.geometry)
        >>> forward_projection = PO.direct(image)
        >>> backward_projection = PO.adjoint(data)

        """

        if acquisition_geometry is None:
            raise TypeError("Please specify an acquisition_geometry to configure this operator")

        if image_geometry == None:
            image_geometry = acquisition_geometry.get_ImageGeometry()

        device = kwargs.get('device', 'gpu')
        if device != 'gpu':
            raise ValueError(
                "TIGRE projectors are GPU only. Got device = {}".format(
                    device))

        DataOrder.check_order_for_engine('tigre', image_geometry)
        DataOrder.check_order_for_engine('tigre', acquisition_geometry)

        super(ProjectionOperator,self).__init__(domain_geometry=image_geometry,\
             range_geometry=acquisition_geometry)

        if direct_method not in ['interpolated', 'Siddon']:
            raise ValueError(
                "direct_method expected 'interpolated' or 'Siddon' got {}".
                format(direct_method))

        if adjoint_weights not in ['matched', 'FDK']:
            raise ValueError(
                "adjoint_weights expected 'matched' or 'FDK' got {}".format(
                    adjoint_weights))

        self.method = {'direct': direct_method, 'adjoint': adjoint_weights}

        #set up TIGRE geometry
        tigre_geom, tigre_angles = CIL2TIGREGeometry.getTIGREGeometry(
            image_geometry, acquisition_geometry)

        tigre_geom.check_geo(tigre_angles)
        tigre_geom.cast_to_single()
        self.tigre_geom = tigre_geom

        #set up TIGRE GPU targets (from 2.2)
        if has_gpu_sel:
            self.gpuids = GpuIds()

    def __call_Ax(self, data):
        if has_gpu_sel:
            return Ax(data, self.tigre_geom, self.tigre_geom.angles,
                      self.method['direct'], self.tigre_geom.mode, self.gpuids)
        else:
            return Ax(data, self.tigre_geom, self.tigre_geom.angles,
                      self.method['direct'], self.tigre_geom.mode)

    def direct(self, x, out=None):

        data = x.as_array()

        if self.tigre_geom.is2D:
            data_temp = np.expand_dims(data, axis=0)
            arr_out = self.__call_Ax(data_temp)
            arr_out = np.squeeze(arr_out, axis=1)
        else:
            arr_out = self.__call_Ax(data)

        #if single angle projection remove the dimension for CIL
        if arr_out.shape[0] == 1:
            arr_out = np.squeeze(arr_out, axis=0)

        if out is None:
            out = AcquisitionData(arr_out,
                                  deep_copy=False,
                                  geometry=self._range_geometry.copy(),
                                  suppress_warning=True)
            return out
        else:
            out.fill(arr_out)

    def __call_Atb(self, data):
        if has_gpu_sel:
            return Atb(data, self.tigre_geom, self.tigre_geom.angles,
                       self.method['adjoint'], self.tigre_geom.mode,
                       self.gpuids)
        else:
            return Atb(data, self.tigre_geom, self.tigre_geom.angles,
                       self.method['adjoint'], self.tigre_geom.mode)

    def adjoint(self, x, out=None):

        data = x.as_array()

        #if single angle projection add the dimension in for TIGRE
        if x.dimension_labels[0] != AcquisitionGeometry.ANGLE:
            data = np.expand_dims(data, axis=0)

        if self.tigre_geom.is2D:
            data = np.expand_dims(data, axis=1)
            arr_out = self.__call_Atb(data)
            arr_out = np.squeeze(arr_out, axis=0)
        else:
            arr_out = self.__call_Atb(data)

        if out is None:
            out = ImageData(arr_out,
                            deep_copy=False,
                            geometry=self._domain_geometry.copy(),
                            suppress_warning=True)
            return out
        else:
            out.fill(arr_out)

    def domain_geometry(self):
        return self._domain_geometry

    def range_geometry(self):
        return self._range_geometry



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.optimisation.algorithms import Algorithm
from cil.optimisation.functions import ZeroFunction
import numpy
import logging
from numbers import Number

log = logging.getLogger(__name__)


class ISTA(Algorithm):

    r"""Iterative Shrinkage-Thresholding Algorithm, see :cite:`BeckTeboulle_b`, :cite:`BeckTeboulle_a`.

    Iterative Shrinkage-Thresholding Algorithm (ISTA)

    .. math:: x^{k+1} = \mathrm{prox}_{\alpha^{k} g}(x^{k} - \alpha^{k}\nabla f(x^{k}))

    is used to solve

    .. math:: \min_{x} f(x) + g(x)

    where :math:`f` is differentiable, :math:`g` has a *simple* proximal operator and :math:`\alpha^{k}`
    is the :code:`step_size` per iteration.

    Note
    ----

    For a constant step size, i.e., :math:`a^{k}=a` for :math:`k\geq1`, convergence of ISTA
    is guaranteed if

    .. math:: \alpha\in(0, \frac{2}{L}),

    where :math:`L` is the Lipschitz constant of :math:`f`, see :cite:`CombettesValerie`.

    Parameters
    ----------

    initial : DataContainer
              Initial guess of ISTA.
    f : Function
        Differentiable function. If `None` is passed, the algorithm will use the ZeroFunction.
    g : Function or `None`
        Convex function with *simple* proximal operator. If `None` is passed, the algorithm will use the ZeroFunction.
    step_size : positive :obj:`float`, default = None
                Step size for the gradient step of ISTA.
                The default :code:`step_size` is :math:`\frac{1}{L}` or 1 if `f=None`.
    kwargs: Keyword arguments
        Arguments from the base class :class:`.Algorithm`.

    Note
    -----
    If the function `g` is set to `None` or to the `ZeroFunction` then the ISTA algorithm is equivalent to Gradient Descent.

    If the function `f` is set to `None` or to the `ZeroFunction` then the ISTA algorithm is equivalent to a Proximal Point Algorithm.

    Examples
    --------

    .. math:: \underset{x}{\mathrm{argmin}}\|A x - b\|^{2}_{2}

    >>> f = LeastSquares(A, b=b, c=0.5)
    >>> g = ZeroFunction()
    >>> ig = Aop.domain
    >>> ista = ISTA(initial = ig.allocate(), f = f, g = g, max_iteration=10)
    >>> ista.run()


    See also
    --------

    :class:`.FISTA`
    :class:`.GD`


    """

    def _provable_convergence_condition(self):
        return self.step_size <= 0.99*2.0/self.f.L

    @property
    def step_size(self):
       return self._step_size

    # Set default step size
    def set_step_size(self, step_size):
        """ Set default step size.
        """

        if step_size is None:
            if isinstance(self.f, ZeroFunction):
                self._step_size = 1

            elif isinstance(self.f.L, Number):
                self._step_size = 0.99*2.0/self.f.L

            else:
                raise ValueError("Function f is not differentiable")

        else:
            self._step_size = step_size

    def __init__(self, initial, f, g, step_size = None, **kwargs):

        super(ISTA, self).__init__(**kwargs)
        self._step_size = None
        self.set_up(initial=initial, f=f, g=g, step_size=step_size, **kwargs)

    def set_up(self, initial, f, g, step_size, **kwargs):
        """Set up of the algorithm"""
        log.info("%s setting up", self.__class__.__name__)
        # set up ISTA
        self.initial = initial
        self.x_old = initial.copy()
        self.x = initial.copy()

        if f is None:
            f = ZeroFunction()

        self.f = f

        if g is None:
            g = ZeroFunction()

        self.g = g

        if isinstance(f, ZeroFunction) and isinstance(g, ZeroFunction):
            raise ValueError('You set both f and g to be the ZeroFunction and thus the iterative method will not update and will remain fixed at the initial value.')

        # set step_size
        self.set_step_size(step_size=step_size)
        self.configured = True
        log.info("%s configured", self.__class__.__name__)


    def update(self):

        r"""Performs a single iteration of ISTA

        .. math:: x_{k+1} = \mathrm{prox}_{\alpha g}(x_{k} - \alpha\nabla f(x_{k}))

        """

        # gradient step
        self.f.gradient(self.x_old, out=self.x)
        self.x_old.sapyb(1., self.x, -self.step_size, out=self.x_old)

        # proximal step
        self.g.proximal(self.x_old, self.step_size, out=self.x)

    def _update_previous_solution(self):
        """ Swaps the references to current and previous solution based on the :func:`~Algorithm.update_previous_solution` of the base class :class:`Algorithm`.
        """
        tmp = self.x_old
        self.x_old = self.x
        self.x = tmp

    def get_output(self):
        " Returns the current solution. "
        return self.x_old

    def update_objective(self):
        """ Updates the objective

        .. math:: f(x) + g(x)

        """
        self.loss.append( self.f(self.x_old) + self.g(self.x_old) )


class FISTA(ISTA):

    r"""Fast Iterative Shrinkage-Thresholding Algorithm, see :cite:`BeckTeboulle_b`, :cite:`BeckTeboulle_a`.

    Fast Iterative Shrinkage-Thresholding Algorithm (FISTA)

    .. math::

        \begin{cases}
            y_{k} = x_{k} - \alpha\nabla f(x_{k})  \\
            x_{k+1} = \mathrm{prox}_{\alpha g}(y_{k})\\
            t_{k+1} = \frac{1+\sqrt{1+ 4t_{k}^{2}}}{2}\\
            y_{k+1} = x_{k} + \frac{t_{k}-1}{t_{k-1}}(x_{k} - x_{k-1})
        \end{cases}

    is used to solve

    .. math:: \min_{x} f(x) + g(x)

    where :math:`f` is differentiable, :math:`g` has a *simple* proximal operator and :math:`\alpha^{k}`
    is the :code:`step_size` per iteration.


    Parameters
    ----------

    initial : DataContainer
            Starting point of the algorithm
    f : Function
        Differentiable function.  If `None` is passed, the algorithm will use the ZeroFunction.
    g : Function or `None`
        Convex function with *simple* proximal operator. If `None` is passed, the algorithm will use the ZeroFunction.
    step_size : positive :obj:`float`, default = None
                Step size for the gradient step of FISTA.
                The default :code:`step_size` is :math:`\frac{1}{L}` or 1 if `f=None`.
    kwargs: Keyword arguments
        Arguments from the base class :class:`.Algorithm`.

    Note
    -----
    If the function `g` is set to `None` or to the `ZeroFunction` then the FISTA algorithm is equivalent to Accelerated Gradient Descent by Nesterov (:cite:`nesterov2003introductory` algorithm 2.2.9).

    If the function `f` is set to `None` or to the `ZeroFunction` then the FISTA algorithm is equivalent to Guler's First Accelerated Proximal Point Method  (:cite:`guler1992new` sec 2).

    Examples
    --------

    .. math:: \underset{x}{\mathrm{argmin}}\|A x - b\|^{2}_{2}


    >>> f = LeastSquares(A, b=b, c=0.5)
    >>> g = ZeroFunction()
    >>> ig = Aop.domain
    >>> fista = FISTA(initial = ig.allocate(), f = f, g = g, max_iteration=10)
    >>> fista.run()

    See also
    --------
    :class:`.FISTA`
    :class:`.GD`

    """

    def set_step_size(self, step_size):

        """Set the default step size
        """

        if step_size is None:

            if isinstance(self.f, ZeroFunction):
                self._step_size = 1

            elif isinstance(self.f.L, Number):
                self._step_size = 1./self.f.L

            else:
                raise ValueError("Function f is not differentiable")

        else:
            self._step_size = step_size

    def _provable_convergence_condition(self):
        return self.step_size <= 1./self.f.L

    def __init__(self, initial, f, g, step_size = None, **kwargs):

        self.y = initial.copy()
        self.t = 1
        super(FISTA, self).__init__(initial=initial, f=f, g=g, step_size=step_size, **kwargs)

    def update(self):

        r"""Performs a single iteration of FISTA

        .. math::

            \begin{cases}
                x_{k+1} = \mathrm{prox}_{\alpha g}(y_{k} - \alpha\nabla f(y_{k}))\\
                t_{k+1} = \frac{1+\sqrt{1+ 4t_{k}^{2}}}{2}\\
                y_{k+1} = x_{k} + \frac{t_{k}-1}{t_{k-1}}(x_{k} - x_{k-1})
            \end{cases}

        """

        self.t_old = self.t

        self.f.gradient(self.y, out=self.x)

        self.y.sapyb(1., self.x, -self.step_size, out=self.y)

        self.g.proximal(self.y, self.step_size, out=self.x)

        self.t = 0.5*(1 + numpy.sqrt(1 + 4*(self.t_old**2)))

        self.x.subtract(self.x_old, out=self.y)
        self.y.sapyb(((self.t_old-1)/self.t), self.x, 1.0, out=self.y)


if __name__ == "__main__":

    from cil.optimisation.functions import L2NormSquared
    from cil.optimisation.algorithms import GD
    from cil.framework import ImageGeometry
    f = L2NormSquared()
    g = L2NormSquared()
    ig = ImageGeometry(3,4,4)
    initial = ig.allocate()
    fista = FISTA(initial, f, g, step_size = 1443432)
    print(fista.is_provably_convergent())

    gd = GD(initial=initial, objective = f, step_size = 1023123)
    print(gd.is_provably_convergent())



#  Copyright 2021 United Kingdom Research and Innovation
#  Copyright 2021 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import AcquisitionGeometry, ImageGeometry
import numpy as np

try:
    from tigre.utilities.geometry import Geometry
except ModuleNotFoundError:
    raise ModuleNotFoundError("This plugin requires the additional package TIGRE\n" +
            "Please install it via conda as tigre from the ccpi channel")

class CIL2TIGREGeometry(object):
    @staticmethod
    def getTIGREGeometry(ig, ag):
        tg = TIGREGeometry(ig, ag)

        #angles
        angles = ag.config.angles.angle_data + ag.config.angles.initial_angle

        if ag.config.angles.angle_unit == AcquisitionGeometry.DEGREE:
            angles *= (np.pi/180.)

        #convert CIL to TIGRE angles s
        angles = -(angles + np.pi/2 +tg.theta )

        #angles in range -pi->pi
        for i, a in enumerate(angles):
            while a < -np.pi:
                a += 2 * np.pi
            while a >= np.pi:
                a -= 2 * np.pi
            angles[i] = a

        return tg, angles

class TIGREGeometry(Geometry):

    def __init__(self, ig, ag):

        Geometry.__init__(self)

        ag_in = ag.copy()
        system = ag_in.config.system
        system.align_reference_frame('tigre')


        #TIGRE's interpolation fp must have the detector outside the reconstruction volume otherwise the ray is clipped
        #https://github.com/CERN/TIGRE/issues/353
        lenx = (ig.voxel_num_x * ig.voxel_size_x)
        leny = (ig.voxel_num_y * ig.voxel_size_y)
        lenz = (ig.voxel_num_z * ig.voxel_size_z)

        panel_width = max(ag_in.config.panel.num_pixels * ag_in.config.panel.pixel_size)*0.5
        clearance_len =  np.sqrt(lenx**2 + leny**2 + lenz**2)/2 + panel_width

        if ag_in.geom_type == 'cone':

            if system.detector.position[1] < clearance_len:

                src = system.source.position.astype(np.float64)
                vec1 = system.detector.position.astype(np.float64) - src

                mag_new = (clearance_len - src[1]) /-src[1]
                scale = mag_new / ag_in.magnification
                scale=np.ceil(scale)

                system.detector.position = src + vec1 * scale
                ag_in.config.panel.pixel_size[0] *= scale
                ag_in.config.panel.pixel_size[1] *= scale

            self.DSO = -system.source.position[1]
            self.DSD = self.DSO + system.detector.position[1]
            self.mode = 'cone'

        else:
            if ag_in.system_description == 'advanced':
                raise NotImplementedError ("CIL cannot use TIGRE to process parallel geometries with tilted axes")

            self.DSO = clearance_len
            self.DSD = 2*clearance_len
            self.mode = 'parallel'

        # number of voxels (vx)
        self.nVoxel = np.array( [ig.voxel_num_z, ig.voxel_num_y, ig.voxel_num_x] )
        # size of each voxel (mm)
        self.dVoxel = np.array( [ig.voxel_size_z, ig.voxel_size_y, ig.voxel_size_x]  )

        # Detector parameters
        # (V,U) number of pixels        (px)
        self.nDetector = np.array(ag_in.config.panel.num_pixels[::-1])
        # size of each pixel            (mm)
        self.dDetector = np.array(ag_in.config.panel.pixel_size[::-1])
        self.sDetector = self.dDetector * self.nDetector    # total size of the detector    (mm)


        if ag_in.dimension == '2D':
            self.is2D = True

            #fix IG to single slice in z
            self.nVoxel[0]=1
            self.dVoxel[0]= ag_in.config.panel.pixel_size[1] / ag_in.magnification

            self.offOrigin = np.array( [0, 0, 0] )

            # Offsets Tigre (Z, Y, X) == CIL (X, -Y)
            if ag_in.geom_type == 'cone':
                self.offDetector = np.array( [0, system.detector.position[0]-system.source.position[0], 0 ])
            else:
                self.offDetector = np.array( [0, system.detector.position[0], 0 ])

            #convert roll, pitch, yaw
            U = [0, system.detector.direction_x[0], -system.detector.direction_x[1]]
            roll = 0
            pitch = 0
            yaw = np.arctan2(-U[2],U[1])

        else:
            self.is2D = False
            # Offsets Tigre (Z, Y, X) == CIL (Z, X, -Y)
            ind = np.asarray([2, 0, 1])
            flip = np.asarray([1, 1, -1])

            if ag_in.geom_type == 'cone':
                #TIGRE origin is at a point on the rotate axis that is in a perpendicular plane containing the source
                #As rotation axis is aligned with z this is the x-y plane
                #We have aligned the y axis with the source->rotate axis direction
                self.offOrigin = np.array( [-system.source.position[2], 0, 0])
                self.offDetector = np.array( [system.detector.position[2]-system.source.position[2], system.detector.position[0]-system.source.position[0], 0])
            else:
                self.offOrigin = np.array( [0,0,0] )
                self.offDetector = np.array( [system.detector.position[2], system.detector.position[0], 0])

            #shift origin z to match image geometry
            #this is in CIL reference frames as the TIGRE geometry rotates the reconstruction volume to match our definitions
            self.offOrigin[0] += ig.center_z


            #convert roll, pitch, yaw
            U = system.detector.direction_x[ind] * flip
            V = system.detector.direction_y[ind] * flip

            roll = np.arctan2(-V[1], V[0])
            pitch = np.arcsin(V[2])
            yaw = np.arctan2(-U[2],U[1])

        #shift origin to match image geometry
        self.offOrigin[1] += ig.center_y
        self.offOrigin[2] += ig.center_x

        self.theta = yaw
        panel_origin = ag_in.config.panel.origin
        if 'right' in panel_origin and 'top' in panel_origin:
            roll += np.pi
        elif 'right' in panel_origin:
            yaw += np.pi
        elif 'top' in panel_origin:
            pitch += np.pi

        self.rotDetector = np.array((roll, pitch, yaw))

        # total size of the image       (mm)
        self.sVoxel = self.nVoxel * self.dVoxel

        # Auxiliary
        self.accuracy = 0.5                        # Accuracy of FWD proj          (vx/sample)



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
from cil.framework import DataContainer, BlockDataContainer
import numpy as np
from utils import initialise_tests

initialise_tests()

def dt(steps):
    return steps[-1] - steps[-2]

class CCPiTestClass(unittest.TestCase):
    def assertBlockDataContainerEqual(self, container1, container2):
        self.assertTrue(issubclass(container1.__class__, container2.__class__))
        for col in range(container1.shape[0]):
            if issubclass(container1.get_item(col).__class__, DataContainer):
                self.assertNumpyArrayEqual(
                    container1.get_item(col).as_array(),
                    container2.get_item(col).as_array()
                    )
            else:
                self.assertBlockDataContainerEqual(container1.get_item(col),container2.get_item(col))


    def assertBlockDataContainerAlmostEqual(self, container1, container2, decimal=7):
        self.assertTrue(issubclass(container1.__class__, container2.__class__))
        for col in range(container1.shape[0]):
            if issubclass(container1.get_item(col).__class__, DataContainer):
                self.assertNumpyArrayAlmostEqual(
                    container1.get_item(col).as_array(),
                    container2.get_item(col).as_array(),
                    decimal=decimal
                    )
            else:
                self.assertBlockDataContainerAlmostEqual(container1.get_item(col),container2.get_item(col), decimal=decimal)


    def assertNumpyArrayEqual(self, first, second):
        np.testing.assert_array_equal(first, second)


    def assertNumpyArrayAlmostEqual(self, first, second, decimal=6):
        np.testing.assert_array_almost_equal(first, second, decimal)



    def assertDataArraysInContainerAllClose(self, container1, container2, rtol=1e-07, msg=None):
        self.assertTrue(issubclass(container1.__class__, container2.__class__))
        if isinstance(container1, BlockDataContainer):
            for col in range(container1.shape[0]):
                if issubclass(container1.get_item(col).__class__, DataContainer):
                    np.testing.assert_allclose(
                        container1.get_item(col).as_array(),
                        container2.get_item(col).as_array(),
                        rtol=rtol,
                        err_msg=msg
                        )
                else:
                    self.assertDataArraysInContainerAllClose(container1.get_item(col),container2.get_item(col), rtol=rtol,  msg=msg)
        else:
            np.testing.assert_allclose(container1.as_array(), container2.as_array(), rtol=rtol, err_msg=msg)



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import numpy
from cil.optimisation.functions import Function
from cil.framework import VectorData, VectorGeometry

class Rosenbrock(Function):
    r'''Rosenbrock function

    .. math::

    F(x,y) = (\alpha - x)^2 + \beta(y-x^2)^2

    The function has a global minimum at .. math:: (x,y)=(\alpha, \alpha^2)

    '''
    def __init__(self, alpha, beta):
        super(Rosenbrock, self).__init__()

        self.alpha = alpha
        self.beta = beta

    def __call__(self, x):
        if not isinstance(x, VectorData):
            raise TypeError('Rosenbrock function works on VectorData only')
        vec = x.as_array()
        a = (self.alpha - vec[0])
        b = (vec[1] - (vec[0]*vec[0]))
        return a * a + self.beta * b * b

    def gradient(self, x, out=None):
        r'''Gradient of the Rosenbrock function

        .. math::

        \nabla f(x,y) = \left[ 2*((x-\alpha) - 2\beta x(y-x^2)) ; 2\beta (y - x^2)  \right]

        '''
        if not isinstance(x, VectorData):
            raise TypeError('Rosenbrock function works on VectorData only')

        vec = x.as_array()
        a = (vec[0] - self.alpha)
        b = (vec[1] - (vec[0]*vec[0]))

        res = numpy.empty_like(vec)
        res[0] = 2 * ( a - 2 * self.beta * vec[0] * b)
        res[1] = 2 * self.beta * b

        if out is not None:
            out.fill (res)
        else:
            return VectorData(res)
